@@ -12,7 +12,7 @@ from stone.ir import (
     unwrap_nullable,
     is_composite_type,
     is_list_type,
-    is_primitive_type,
+    is_map_type,
     is_struct_type,
     Void,
 )
@@ -72,9 +72,13 @@ def _rename_if_reserved(s):
 def fmt_type(data_type, namespace=None, use_interface=False, raw=False):
     data_type, nullable = unwrap_nullable(data_type)
     if is_list_type(data_type):
-        if raw and is_primitive_type(data_type.data_type):
+        if raw and not _needs_base_type(data_type.data_type):
             return "json.RawMessage"
         return '[]%s' % fmt_type(data_type.data_type, namespace, use_interface, raw)
+    if is_map_type(data_type):
+        if raw and not _needs_base_type(data_type.data_type):
+            return "json.RawMessage"
+        return 'map[string]%s' % fmt_type(data_type.value_data_type, namespace, use_interface, raw)
     if raw:
         return "json.RawMessage"
     type_name = data_type.name
@@ -86,7 +90,12 @@ def fmt_type(data_type, namespace=None, use_interface=False, raw=False):
     if use_interface and _needs_base_type(data_type):
         return _type_table.get(data_type.__class__, type_name)
     else:
-        return _type_table.get(data_type.__class__, '*' + type_name)
+        if data_type.__class__ not in _type_table:
+            return '*' + type_name
+        if data_type.__class__ == Timestamp:
+            # For other primitive types, `omitempty` does the job.
+            return ('*' if nullable else '') + _type_table[data_type.__class__]
+        return _type_table[data_type.__class__]
 
 
 def fmt_var(name, export=True, check_reserved=False):
@@ -137,10 +146,13 @@ def generate_doc(code_generator, t):
 
 
 def _needs_base_type(data_type):
+    data_type, _ = unwrap_nullable(data_type)
     if is_struct_type(data_type) and data_type.has_enumerated_subtypes():
         return True
     if is_list_type(data_type):
         return _needs_base_type(data_type.data_type)
+    if is_map_type(data_type):
+        return _needs_base_type(data_type.value_data_type)
     return False
 
 
