@@ -21,13 +21,17 @@
 package dropbox
 
 import (
+	"bytes"
+	"context"
 	"encoding/json"
+	"errors"
 	"fmt"
 	"io"
+	"io/ioutil"
 	"log"
 	"net/http"
+	"strings"
 
-	"golang.org/x/net/context"
 	"golang.org/x/oauth2"
 )
 
@@ -46,6 +50,29 @@ func Version() (string, string) {
 	return sdkVersion, specVersion
 }
 
+// Tagged is used for tagged unions.
+type Tagged struct {
+	Tag string `json:".tag"`
+}
+
+// APIError is the base type for endpoint-specific errors.
+type APIError struct {
+	ErrorSummary string `json:"error_summary"`
+}
+
+func (e APIError) Error() string {
+	return e.ErrorSummary
+}
+
+type SDKInternalError struct {
+	StatusCode int
+	Content    string
+}
+
+func (e SDKInternalError) Error() string {
+	return fmt.Sprintf("Unexpected error: %v (code: %v)", e.Content, e.StatusCode)
+}
+
 // Config contains parameters for configuring the SDK.
 type Config struct {
 	// OAuth2 access token
@@ -56,14 +83,18 @@ type Config struct {
 	Logger *log.Logger
 	// Used with APIs that support operations as another user
 	AsMemberID string
+	// Used with APIs that support operations as an admin
+	AsAdminID string
+	// Path relative to which action should be taken
+	PathRoot string
 	// No need to set -- for testing only
 	Domain string
 	// No need to set -- for testing only
 	Client *http.Client
 	// No need to set -- for testing only
-	HeaderGenerator func(hostType string, style string, namespace string, route string) map[string]string
+	HeaderGenerator func(hostType string, namespace string, route string) map[string]string
 	// No need to set -- for testing only
-	URLGenerator func(hostType string, style string, namespace string, route string) string
+	URLGenerator func(hostType string, namespace string, route string) string
 }
 
 // LogLevel defines a type that can set the desired level of logging the SDK will generate.
@@ -105,42 +136,130 @@ func (c *Config) LogInfo(format string, v ...interface{}) {
 	c.doLog(LogInfo, format, v...)
 }
 
+// Ergonomic methods to set namespace relative to which action should be taken
+func (c Config) WithNamespaceID(nsID string) Config {
+	c.PathRoot = fmt.Sprintf(`{".tag": "namespace_id", "namespace_id": "%s"}`, nsID)
+	return c
+}
+
+func (c Config) WithRoot(nsID string) Config {
+	c.PathRoot = fmt.Sprintf(`{".tag": "root", "root": "%s"}`, nsID)
+	return c
+}
+
 // Context is the base client context used to implement per-namespace clients.
 type Context struct {
 	Config          Config
 	Client          *http.Client
-	HeaderGenerator func(hostType string, style string, namespace string, route string) map[string]string
-	URLGenerator    func(hostType string, style string, namespace string, route string) string
+	NoAuthClient    *http.Client
+	HeaderGenerator func(hostType string, namespace string, route string) map[string]string
+	URLGenerator    func(hostType string, namespace string, route string) string
 }
 
-// NewRequest returns an appropriate Request object for the given namespace/route.
-func (c *Context) NewRequest(
-	hostType string,
-	style string,
-	authed bool,
-	namespace string,
-	route string,
-	headers map[string]string,
-	body io.Reader,
-) (*http.Request, error) {
-	url := c.URLGenerator(hostType, style, namespace, route)
-	req, err := http.NewRequest("POST", url, body)
+type Request struct {
+	Host      string
+	Namespace string
+	Route     string
+	Style     string
+	Auth      string
+
+	Arg          interface{}
+	ExtraHeaders map[string]string
+}
+
+func (c *Context) Execute(req Request, body io.Reader) ([]byte, io.ReadCloser, error) {
+	url := c.URLGenerator(req.Host, req.Namespace, req.Route)
+	httpReq, err := http.NewRequest("POST", url, body)
 	if err != nil {
-		return nil, err
+		return nil, nil, err
 	}
-	for k, v := range headers {
-		req.Header.Add(k, v)
+
+	for k, v := range req.ExtraHeaders {
+		httpReq.Header.Add(k, v)
 	}
-	for k, v := range c.HeaderGenerator(hostType, style, namespace, route) {
-		req.Header.Add(k, v)
+
+	for k, v := range c.HeaderGenerator(req.Host, req.Namespace, req.Route) {
+		httpReq.Header.Add(k, v)
 	}
-	if req.Header.Get("Host") != "" {
-		req.Host = req.Header.Get("Host")
+
+	if httpReq.Header.Get("Host") != "" {
+		httpReq.Host = httpReq.Header.Get("Host")
 	}
-	if !authed {
-		req.Header.Del("Authorization")
+
+	if req.Auth == "noauth" {
+		httpReq.Header.Del("Authorization")
+	}
+	if req.Auth != "team" && c.Config.AsMemberID != "" {
+		httpReq.Header.Add("Dropbox-API-Select-User", c.Config.AsMemberID)
+	}
+	if req.Auth != "team" && c.Config.AsAdminID != "" {
+		httpReq.Header.Add("Dropbox-API-Select-Admin", c.Config.AsAdminID)
+	}
+	if c.Config.PathRoot != "" {
+		httpReq.Header.Add("Dropbox-API-Path-Root", c.Config.PathRoot)
+	}
+
+	if req.Arg != nil {
+		serializedArg, err := json.Marshal(req.Arg)
+		if err != nil {
+			return nil, nil, err
+		}
+
+		switch req.Style {
+		case "rpc":
+			if body != nil {
+				return nil, nil, errors.New("RPC style requests can not have body")
+			}
+
+			httpReq.Header.Set("Content-Type", "application/json")
+			httpReq.Body = ioutil.NopCloser(bytes.NewReader(serializedArg))
+			httpReq.ContentLength = int64(len(serializedArg))
+		case "upload", "download":
+			httpReq.Header.Set("Dropbox-API-Arg", string(serializedArg))
+			httpReq.Header.Set("Content-Type", "application/octet-stream")
+		}
+	}
+
+	client := c.Client
+	if req.Auth == "noauth" {
+		client = c.NoAuthClient
+	}
+
+	resp, err := client.Do(httpReq)
+	if err != nil {
+		return nil, nil, err
+	}
+
+	if resp.StatusCode == http.StatusOK || resp.StatusCode == http.StatusPartialContent {
+		switch req.Style {
+		case "rpc", "upload":
+			if resp.Body == nil {
+				return nil, nil, errors.New("Expected body in RPC response, got nil")
+			}
+
+			b, err := ioutil.ReadAll(resp.Body)
+			resp.Body.Close()
+			if err != nil {
+				return nil, nil, err
+			}
+
+			return b, nil, nil
+		case "download":
+			b := []byte(resp.Header.Get("Dropbox-API-Result"))
+			return b, resp.Body, nil
+		}
+	}
+
+	b, err := ioutil.ReadAll(resp.Body)
+	resp.Body.Close()
+	if err != nil {
+		return nil, nil, err
+	}
+
+	return nil, nil, SDKInternalError{
+		StatusCode: resp.StatusCode,
+		Content:    string(b),
 	}
-	return req, nil
 }
 
 // NewContext returns a new Context with the given Config.
@@ -157,9 +276,14 @@ func NewContext(c Config) Context {
 		client = conf.Client(context.Background(), tok)
 	}
 
+	noAuthClient := c.Client
+	if noAuthClient == nil {
+		noAuthClient = &http.Client{}
+	}
+
 	headerGenerator := c.HeaderGenerator
 	if headerGenerator == nil {
-		headerGenerator = func(hostType string, style string, namespace string, route string) map[string]string {
+		headerGenerator = func(hostType string, namespace string, route string) map[string]string {
 			return map[string]string{}
 		}
 	}
@@ -171,13 +295,13 @@ func NewContext(c Config) Context {
 			hostContent: hostContent + domain,
 			hostNotify:  hostNotify + domain,
 		}
-		urlGenerator = func(hostType string, style string, namespace string, route string) string {
+		urlGenerator = func(hostType string, namespace string, route string) string {
 			fqHost := hostMap[hostType]
 			return fmt.Sprintf("https://%s/%d/%s/%s", fqHost, apiVersion, namespace, route)
 		}
 	}
 
-	return Context{c, client, headerGenerator, urlGenerator}
+	return Context{c, client, noAuthClient, headerGenerator, urlGenerator}
 }
 
 // OAuthEndpoint constructs an `oauth2.Endpoint` for the given domain
@@ -193,31 +317,19 @@ func OAuthEndpoint(domain string) oauth2.Endpoint {
 	return oauth2.Endpoint{AuthURL: authURL, TokenURL: tokenURL}
 }
 
-// Tagged is used for tagged unions.
-type Tagged struct {
-	Tag string `json:".tag"`
-}
-
-// APIError is the base type for endpoint-specific errors.
-type APIError struct {
-	ErrorSummary string `json:"error_summary"`
-}
-
-func (e APIError) Error() string {
-	return e.ErrorSummary
-}
-
-// HandleCommonAPIErrors handles common API errors
-func HandleCommonAPIErrors(c Config, resp *http.Response, body []byte) error {
-	var apiError APIError
-	if resp.StatusCode == http.StatusBadRequest || resp.StatusCode == http.StatusInternalServerError {
-		apiError.ErrorSummary = string(body)
-		return apiError
+// HTTPHeaderSafeJSON encode the JSON passed in b []byte passed in in
+// a way that is suitable for HTTP headers.
+//
+// See: https://www.dropbox.com/developers/reference/json-encoding
+func HTTPHeaderSafeJSON(b []byte) string {
+	var s strings.Builder
+	s.Grow(len(b))
+	for _, r := range string(b) {
+		if r >= 0x007f {
+			fmt.Fprintf(&s, "\\u%04x", r)
+		} else {
+			s.WriteRune(r)
+		}
 	}
-	e := json.Unmarshal(body, &apiError)
-	if e != nil {
-		c.LogDebug("%v", e)
-		return e
-	}
-	return apiError
+	return s.String()
 }
